import re, math, copy
from enum import Enum

from coldtype.time import Timeable, Frame
from coldtype.time.easing import ease
from coldtype.time.timeline import Timeline

from coldtype.text import StyledString, Lockup, Graf, GrafStyle, Style
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.color import *

from coldtype.time.clip import Clip, ClipFlags, ClipType

from typing import Optional, Union, Callable, Tuple


class Marker(Timeable):
    def __init__(self, start, end, marker):
        # TODO super name= marker name?
        super().__init__(start, end, data=marker)


class ClipGroupPens(DATPens):
    def __init__(self, clip_group):
        super().__init__()
        self.cg = clip_group
    
    def _iterate_tags(self, tag, pens):
        if hasattr(pens, "_pens"):
            for pen in pens:
                if pen._tag == tag:
                    yield pen.data.get(tag), pen
                else:
                    yield from self._iterate_tags(tag, pen)
    
    def iterate_clips(self):
        yield from self._iterate_tags("clip", self)
    
    def iterate_slugs(self):
        yield from self._iterate_tags("slug", self)
    
    def iterate_lines(self):
        yield from self._iterate_tags("line", self)
    
    def map_clips(self, fn):
        for clip, pen in self.iterate_clips():
            fn(clip, pen)
        return self
    
    def clean_empties(self):
        for _, pen in self.iterate_slugs():
            pen._pens = [p for p in pen._pens if len(p._pens) > 0]
        for line in self:
            line._pens = [p for p in line._pens if len(p._pens) > 0]
        self._pens = [p for p in self._pens if len(p._pens) > 0]
        return self

    def remove_futures(self, clean=True):
        for clip, pen in self.iterate_clips():
            if clip.position > 0:
                pen._pens = []
        
        if clean:
            self.clean_empties()
        return self
    
    def remove_future_lines(self, clean=True):
        for _, line in self.iterate_lines():
            any_now = False
            for pen in line:
                for p in pen:
                    clip = p.data.get("clip")
                    if clip:
                        if clip.position <= 0:
                            any_now = True
            
            if not any_now:
                line._pens = []
            #for x in line.iterate_clips():
            #    print(">>>", x)

        #for clip, pen in self.iterate_clips():
        #    if clip.position > 0:
        #        pen._pens = []
        
        if clean:
            self.clean_empties()
        return self
    
    def understroke(self, s=0, sw=5):
        for clip, pen in self.iterate_clips():
            pen.understroke(s=s, sw=sw)
        return self


class ClipGroup(Timeable):
    def __init__(self, timeline, index, clips):
        self.index = index
        self.clips = clips
        self.timeline = timeline
        self.style_indices = []

        self.retime()

        for idx, clip in enumerate(clips):
            clip.idx = idx
            clip.group = self
    
    def retime(self):
        if len(self.clips) > 0:
            self.start = self.clips[0].start
            self.end = self.clips[-1].end
            self.track = self.clips[0].track
            self.valid = True
        else:
            self.start = 0
            self.end = 0
            self.track = None
            self.valid = False
    
    def styles(self):
        all_styles = set()
        for clip in self.clips:
            for style in clip.styles:
                all_styles.add(style)
        return all_styles
    
    def style_matching(self, style):
        for clip in self.clips:
            match = clip.style_matching(style)
            if match:
                return match
    
    def ldata(self, field, default):
        if len(self.clips) > 0:
            return self.clips[-1].inline_data.get(field, default)
    
    def current_style_matching(self, f:Frame, identifier:Union[str, Callable[[Clip], bool]]):
        for si in self.style_indices:
            sc = self.timeline.sequence[si].current(f.i)
            if sc and identifier(sc):
                if isinstance(identifier, str):
                    if identifier in sc.text:
                        return sc
                elif identifier(sc):
                    return sc
    
    def lines(self, ignore_newlines=False):
        lines = []
        line = []
        for clip in self.clips:
            if clip.type == ClipType.NewLine:
                if ignore_newlines:
                    if not clip.text.startswith(" "):
                        clip.text = " " + clip.text
                    line.append(clip)
                else:
                    lines.append(line)
                    line = [clip]
            elif clip.type == ClipType.GrafBreak:
                lines.append(line)
                cclip = copy.deepcopy(clip)
                cclip.text = "¶"
                lines.append([cclip])
                line = [clip]
                # add a graf mark
            else:
                line.append(clip)
        if len(line) > 0:
            lines.append(line)
        return lines
    
    def position(self, idx, styles):
        for clip in self.clips:
            clip.joined = False
        for clip in self.clips:
            clip:Clip
            clip.styles = []
            if False:
                if styles:
                    for style in styles:
                        if style:
                            for style_clip in style.clips:
                                if clip.start >= style_clip.start and clip.end <= style_clip.end:
                                    clip.styles.append(style_clip.ftext().strip())
                                elif style_clip.start >= clip.start and style_clip.end <= clip.end:
                                    clip.styles.append(style_clip.ftext().strip())
            if clip.start > idx:
                clip.position = 1
            elif clip.start <= idx and clip.end > idx:
                clip.position = 0
                before_clip = clip
                while before_clip.type == ClipType.JoinPrev:
                    before_clip = self.clips[before_clip.idx-1]
                    before_clip.joined = True
                try:
                    after_clip = self.clips[clip.idx+1]
                    while after_clip.type == ClipType.JoinPrev:
                        after_clip.joined = True
                        after_clip = self.clips[after_clip.idx+1]
                except IndexError:
                    pass
            else:
                clip.position = -1
            
            if True:
                for style in styles:
                    style:ClipTrack
                    if clip.position == -1:
                        sc = style.current(clip.end-1)
                    elif clip.position == 0:
                        sc = style.current(idx)
                    else:
                        sc = style.current(clip.start)
                    if sc:
                        clip.styles.extend([s.strip() for s in sc.ftext().strip().split(",")])
                        clip.style_clips.append(sc)
        return self
    
    def currentSyllable(self):
        for clip in self.clips:
            if clip.position == 0:
                return clip
    
    def currentWord(self):
        for clip in self.clips:
            if clip.position == 0:
                clips = [clip]
                before_clip = clip
                # walk back
                while before_clip.type == ClipType.JoinPrev:
                    before_clip = self.clips[before_clip.idx-1]
                    clips.insert(0, before_clip)
                # walk forward
                try:
                    after_clip = self.clips[clip.idx+1]
                    while after_clip.type == ClipType.JoinPrev:
                        clips.append(after_clip)
                        after_clip = self.clips[after_clip.idx+1]
                except IndexError:
                    pass
                return clips
    
    def currentLine(self):
        for line in self.lines():
            for clip in line:
                if clip.position == 0:
                    return line

    def sibling(self, clip, direction, wrap=False):
        try:
            if clip.idx + direction < 0 and not wrap:
                return None
            elif clip.idx + direction >= len(self.clips):
                if not wrap:
                    return None
                else:
                    return self.clips[0]
            return self.clips[clip.idx + direction]
        except IndexError:
            return None
    
    def text(self):
        txt = ""
        for c in self.clips:
            if c.type == ClipType.ClearScreen:
                pass
            elif c.type == ClipType.Isolated:
                txt += "( )"
            elif c.type == ClipType.JoinPrev:
                txt += "|"
            elif c.type == ClipType.NewLine:
                txt += "/(\\n)/"
            txt += c.text
        return txt
    
    def pens(self,
        f,
        render_clip_fn:Callable[[Frame, int, Clip, str], Tuple[str, Style]],
        rect=None,
        graf_style=GrafStyle(leading=20),
        fit=None,
        ignore_newlines=False,
        removeOverlap=False) -> ClipGroupPens:
        """
        render_clip_fn: frame, line index, clip, clip text — you must return a tuple of text to render and the Style to render it with
        """
        if not rect:
            rect = f.a.r
        group_pens = ClipGroupPens(self)
        lines = []
        groupings = []
        for idx, _line in enumerate(self.lines(ignore_newlines=ignore_newlines)):
            slugs = []
            texts = []
            for clip in _line:
                if clip.type == ClipType.Meta or clip.blank:
                    continue
                try:
                    ftext = clip.ftext()
                    text, style = render_clip_fn(f, idx, clip, ftext)
                    texts.append([text, clip.idx, style])
                except Exception as e:
                    print(e)
                    raise Exception("pens render_clip must return text & style")
            
            grouped_texts = []
            idx = 0
            done = False
            while not done:
                style = texts[idx][2]
                grouped_text = [texts[idx]]
                style_same = True
                while style_same:
                    idx += 1
                    try:
                        next_style = texts[idx][2]
                        if next_style == style:
                            style_same = True
                            grouped_text.append(texts[idx])
                        else:
                            style_same = False
                            grouped_texts.append(grouped_text)
                    except IndexError:
                        done = True
                        style_same = False
                        grouped_texts.append(grouped_text)
            
            for gt in grouped_texts:
                full_text = "".join([t[0] for t in gt])
                slugs.append(StyledString(full_text, gt[0][2]))
            groupings.append(grouped_texts)
            lines.append(slugs)
        
        lockups = []
        for line in lines:
            lockup = Lockup(line, preserveLetters=True, nestSlugs=True)
            if fit:
                lockup.fit(fit)
            lockups.append(lockup)
        graf = Graf(lockups, rect, graf_style)
        pens = graf.pens()#.align(rect, x="minx")
        
        re_grouped = DATPens()
        for idx, line in enumerate(lines):
            #print(pens, idx, line[0].text)
            line_dps = pens[idx]
            re_grouped_line = DATPens()
            re_grouped_line.tag("line")
            position = 1
            line_text = ""
            for gidx, gt in enumerate(groupings[idx]):
                group_dps = line_dps[gidx]
                tidx = 0
                last_clip_dps = None
                for (text, cidx, style) in gt:
                    clip:Clip = self.clips[cidx]
                    if clip.position == 0:
                        position = 0
                    elif clip.position == -1:
                        position = -1
                    line_text += clip.ftext()
                    clip_dps = DATPens(group_dps[tidx:tidx+len(text)])
                    clip_dps.tag("clip")
                    clip_dps.data["clip"] = self.clips[cidx]
                    clip_dps.data["line_index"] = idx
                    clip_dps.data["line"] = re_grouped_line
                    clip_dps.data["group"] = re_grouped
                    if clip.type == ClipType.JoinPrev and last_clip_dps:
                        grouped_clip_dps = last_clip_dps #DATPens()
                        #grouped_clip_dps.append(last_clip_dps)
                        #grouped_clip_dps._pens = last_clip_dps._pens
                        grouped_clip_dps.append(clip_dps)
                        #re_grouped_line[-1] = grouped_clip_dps
                        #last_clip_dps = grouped_clip_dps
                    else:
                        last_clip_dps = DATPens()
                        last_clip_dps.tag("slug")
                        last_clip_dps.append(clip_dps)
                        re_grouped_line.append(last_clip_dps)
                    tidx += len(text)
            re_grouped_line.data["line_index"] = idx
            re_grouped_line.data["position"] = position
            re_grouped_line.data["line_text"] = line_text
            re_grouped_line.addFrame(line_dps.getFrame())
            re_grouped.append(re_grouped_line)
        
        pens = re_grouped
        for pens in pens._pens:
            if removeOverlap:
                for pen in pens._pens:
                    pen.removeOverlap()
            group_pens.append(pens)
        
        for clip, pen in group_pens.iterate_clips():
            if clip.position == 1 or clip.text == "¶":
                #pen.f(0)
                pass
            if clip.blank:
                pen._pens = []
        
        for line in group_pens:
            line.reversePens() # line top-to-bottom
            for slug in line:
                slug.reversePens() # slugs left-to-right
                for clip in slug:
                    clip.reversePens() # glyphs left-to-right #.understroke(sw=5)
        
        group_pens.reversePens()
        return group_pens
    
    def iterate_clip_pens(self, pens):
        if hasattr(pens, "_pens"):
            for pen in pens:
                if hasattr(pen, "data"):
                    if pen.data.get("clip"):
                        yield pen.data.get("clip"), pen
                    else:
                        yield from self.iterate_clip_pens(pen)

    def remove_futures(self, pens):
        for clip, pen in self.iterate_clip_pens(pens):
            if clip.position > 0:
                pen._pens = []
    
    def iterate_pens(self, pens, copy=True):
        for idx, line in enumerate(self.lines()):
            _pens = pens[idx]
            for cidx, clip in enumerate(line):
                if copy:
                    p = _pens._pens[cidx].copy()
                else:
                    p = _pens._pens[cidx]
                yield idx, clip, p
    
    def __repr__(self):
        return "<ClipGroup {:04d}-{:04d} \"{:s}\">".format(self.start, self.end, self.text())
    
    def __hash__(self):
        return self.text()


class ClipTrack():
    def __init__(self, sequence, clips, markers):
        self.sequence = sequence
        self.clips = clips
        self.markers = markers
        self.clip_groups = self.groupedClips(clips)
    
    def groupedClips(self, clips):
        groups = []
        group = []
        last_clip = None
        for idx, clip in enumerate(clips):
            if clip.type == ClipType.ClearScreen:
                if len(group) > 0:
                    groups.append(ClipGroup(self, len(groups), group))
                group = []
            elif clip.type == ClipType.JoinPrev:
                if last_clip:
                    last_clip.addJoin(clip, +1)
                    clip.addJoin(last_clip, -1)
            group.append(clip)
            last_clip = clip
        if len(group) > 0:
            groups.append(ClipGroup(self, len(groups), group))
        return groups
    
    def current(self, fi):
        for idx, clip in enumerate(self.clips):
            clip:Clip
            if clip.start <= fi and fi < clip.end:
                return clip
    
    def now(self, fi, text):
        for idx, clip in enumerate(self.clips):
            clip:Clip
            if clip.text == text and clip.start <= fi and fi < clip.end:
                return clip
        return False

    def now_or_past(self, fi, text):
        for idx, clip in enumerate(self.clips):
            clip:Clip
            if clip.text == text and clip.start <= fi:
                return clip
        return False
    
    def over(self, fi, text):
        for idx, clip in enumerate(self.clips):
            clip:Clip
            if clip.text == text and fi >= clip.end:
                return True
        return False
    
    def duration(self):
        duration = 0
        for c in self.clips:
            duration = max(duration, c.end)
        return duration
    
    def __repr__(self):
        return "<ClipTrack {:s}>".format("/".join([c.ftext() for c in self.clips[:3]]))


class Sequence(Timeline):
    def __init__(self, duration, fps, storyboard, tracks, workarea_track=0):
        self.workarea_track = workarea_track
        super().__init__(duration, fps, storyboard, tracks)
    
    def find_symbol(self, symbol):
        start = -1
        end = -1
        for t in self.tracks:
            for c in t.clips:
                c:Clip
                if c.symbol and c.symbol == symbol:
                    if c.symbol_position == -2:
                        start = c.start
                        end = c.end
                    elif c.symbol_position == -1:
                        start = c.start
                    elif c.symbol_position == +1:
                        end = c.end
        return start, end
    
    def retime_for_symbol(self, symbol):
        start, end = self.find_symbol(symbol)
        if start == -1 or end == -1:
            raise Exception("No symbol found")
        
        duration = end - start
        self.start = 0
        self.end = duration

        # TODO storyboard?
        for idx, frame in enumerate(self.storyboard):
            self.storyboard[idx] = frame - start

        for t in self.tracks:
            t:ClipTrack
            for c in t.clips:
                c:Clip
                c.start = c.start - start
                c.end = c.end - start
            
            for cg in t.clip_groups:
                cg.retime()
        
        return self
    
    def trackClipGroupForFrame(self, track_idx, frame_idx, styles=[], check_end=True):
        for gidx, group in enumerate(self[track_idx].clip_groups):
            group:ClipGroup
            group.style_indices = styles
            if not check_end:
                end_good = False
                try:
                    next_group = self[track_idx].clip_groups[gidx+1]
                    end_good = next_group.start > frame_idx
                except IndexError:
                    end_good = True
            if group.start <= frame_idx and ((check_end and group.end > frame_idx) or (not check_end and end_good)):
                if True:
                    style_tracks = [self[sidx] for sidx in styles]
                if False:
                    style_groups = None
                    if styles:
                        style_groups = []
                        for style in styles:
                            style_groups.append(self.trackClipGroupForFrame(style, frame_idx, check_end=False))
                return group.position(frame_idx, style_tracks)
    
    def clip_group(self, track_idx, f, styles=[]) -> ClipGroup:
        cg = self.trackClipGroupForFrame(track_idx, f.i if hasattr(f, "i") else f, styles)
        if cg:
            return cg
        else:
            return ClipGroup(self[track_idx], -1, [])
    
    def find_workarea(self):
        cg = self.trackClipGroupForFrame(self.workarea_track, self.cti)
        if cg:
            return [cg.start, cg.end]
    
    def jumps(self):
        js = self._jumps
        t = self[self.workarea_track]
        for clip in t.clips:
            js.insert(-1, clip.start)
        return js
    
    def text_for_frame(self, fi):
        cg = self.clip_group(self.workarea_track, fi)
        if cg and cg.currentSyllable():
            cgs = cg.currentSyllable()
            if cgs.start == fi:
                return cgs.input_text


from enum import Enum, auto

try:
    import skia
except ImportError:
    skia = None

class BlendMode(Enum):
    Clear = auto()
    Src = auto()
    Dst = auto()
    SrcOver = auto()
    DstOver = auto()
    SrcIn = auto()
    DstIn = auto()
    SrcOut = auto()
    DstOut = auto()
    SrcATop = auto()
    DstATop = auto()
    Xor = auto()
    Plus = auto()
    Modulate = auto()
    Screen = auto()
    Overlay = auto()
    Darken = auto()
    Lighten = auto()
    ColorDodge = auto()
    ColorBurn = auto()
    HardLight = auto()
    SoftLight = auto()
    Difference = auto()
    Exclusion = auto()
    Multiply = auto()
    Hue = auto()
    Saturation = auto()
    Color = auto()
    Luminosity = auto()
    
    def to_skia(self):
        return _SKIA_MAPPING[self.value-1]

if not skia:
    _SKIA_MAPPING = []
else:
    _SKIA_MAPPING = [
        skia.BlendMode.kClear,
        skia.BlendMode.kSrc,
        skia.BlendMode.kDst,
        skia.BlendMode.kSrcOver,
        skia.BlendMode.kDstOver,
        skia.BlendMode.kSrcIn,
        skia.BlendMode.kDstIn,
        skia.BlendMode.kSrcOut,
        skia.BlendMode.kDstOut,
        skia.BlendMode.kSrcATop,
        skia.BlendMode.kDstATop,
        skia.BlendMode.kXor,
        skia.BlendMode.kPlus,
        skia.BlendMode.kModulate,
        skia.BlendMode.kScreen,
        skia.BlendMode.kOverlay,
        skia.BlendMode.kDarken,
        skia.BlendMode.kLighten,
        skia.BlendMode.kColorDodge,
        skia.BlendMode.kColorBurn,
        skia.BlendMode.kHardLight,
        skia.BlendMode.kSoftLight,
        skia.BlendMode.kDifference,
        skia.BlendMode.kExclusion,
        skia.BlendMode.kMultiply,
        skia.BlendMode.kHue,
        skia.BlendMode.kSaturation,
        skia.BlendMode.kColor,
        skia.BlendMode.kLuminosity]

if __name__ == "__main__":
    print(BlendMode.Saturation.to_skia())


import ast, json
from http.server import SimpleHTTPRequestHandler
from enum import Enum

try:
    from SimpleWebSocketServer import SimpleWebSocketServer, WebSocket
except ImportError:
    SimpleWebSocketServer = object
    WebSocket = object


class Watchable(Enum):
    Source = "Source"
    Font = "Font"
    Library = "Library"
    Generic = "Generic"
    

class EditAction(Enum):
    Newline = "newline"
    Tokenize = "tokenize"
    TokenizeLine = "tokenize_line"
    NewSection = "newsection"
    Capitalize = "capitalize"
    SelectWorkarea = "select_workarea"


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class SimpleEcho(WebSocket):
    def handleMessage(self):
        if "webviewer" in self.data:
            data = json.loads(self.data)
            if data.get("webviewer") == True:
                self.webviewer = True
        #print("INCOMING!", self, self.address, self.data)
        self.messages.append(self.data)

    def handleConnected(self):
        #if not str(self.address).startswith("('::ffff"):
        #    print(self.address, "connected")
        self.messages = []
        self.webviewer = False

    def handleClose(self):
        #print(self.address, "closed")
        pass


def echo_server(port):
    return SimpleWebSocketServer('', port, SimpleEcho)


def bytesto(bytes):
    r = float(bytes)
    for i in range(2):
        r = r / 1024
    return(r)


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def file_and_line_to_def(filepath, lineno):
    # https://julien.danjou.info/finding-definitions-from-a-source-file-and-a-line-number-in-python/
    candidate = None
    for item in ast.walk(ast.parse(filepath.read_text())):
        if isinstance(item, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            if item.lineno > lineno:
                continue
            if candidate:
                distance = lineno - item.lineno
                if distance < (lineno - candidate.lineno):
                    candidate = item
            else:
                candidate = item
    if candidate:
        return candidate.name


import enum
import math, re

from time import sleep
from typing import Callable, Optional, Tuple, Type
from collections import OrderedDict
from copy import deepcopy

from fontTools.misc.transform import Transform
from fontTools.pens.boundsPen import BoundsPen
from fontTools.pens.recordingPen import RecordingPen
from fontTools.pens.transformPen import TransformPen
from fontTools.pens.reverseContourPen import ReverseContourPen

from fontPens.flattenPen import FlattenPen
from coldtype.geometry import Atom, Point, Line, Edge, Rect, Curve, align
from coldtype.color import normalize_color
from coldtype.sh import SH_UNARY_SUFFIX_PROPS, sh, SHContext

from coldtype.pens.misc import BooleanOp, calculate_pathop, ExplodingPen, SmoothPointsPen

from coldtype.pens.outlinepen import OutlinePen
from coldtype.pens.translationpen import TranslationPen, polarCoord

from coldtype.beziers import CurveCutter, splitCubicAtT
from coldtype.interpolation import norm
from coldtype.grid import Grid


class DraftingPen(RecordingPen, SHContext):
    """Fluent subclass of RecordingPen"""

    def __init__(self, *args):
        SHContext.__init__(self)
        RecordingPen.__init__(self)

        from coldtype.pens.draftingpens import DraftingPens

        self.single_pen_class = DraftingPen
        self.multi_pen_class = DraftingPens

        self._tag = None
        self._frame = None
        self._visible = True
        self._parent = None
        self._last = None
        self._alpha = 1

        self._typographic = False
        self.glyphName = None

        self._current_attr_tag = "default"
        self.clearAttrs()
        self.data = {}

        self.defs = None
        self.macros = {}

        for _, arg in enumerate(args):
            if isinstance(arg, str):
                self.tag(arg)
            elif isinstance(arg, DraftingPen):
                self.replace_with(arg)
            elif isinstance(arg, Rect):
                self.rect(arg)
            elif isinstance(arg, Line):
                self.line(arg)
            elif isinstance(arg, Point):
                self.oval(Rect.FromCenter(arg, 50, 50))
    
    def __repr__(self):
        s = f"{type(self).__name__}<"
        if self._tag:
            s += self._tag + ":"
        if len(self.value) == 0:
            s += "(((empty)))"
        else:
            s += f"{len(self.value)}mvs:"
            if self.value[-1][0] == "closePath":
                s += "closed"
            elif self.value[-1][0] == "endPath":
                s += "end"
            else:
                s += "open"
        s += "/>"
        return s
    
    def tag(self, value=None):
        if value:
            if isinstance(value, str):
                self._tag = value
            return self
        else:
            return self._tag
        
    def frame(self, value=None):
        if value:
            if isinstance(value, Rect):
                self._frame = value
            return self
        else:
            return self._frame
    
    def visible(self, value=None):
        if value is not None:
            self._visible = value
            return self
        else:
            return self._visible
        
    def bounds(self):
        """Calculate the bounds of this shape; mostly for internal use."""
        try:
            cbp = BoundsPen(None)
            self.replay(cbp)
            mnx, mny, mxx, mxy = cbp.bounds
            return Rect((mnx, mny, mxx - mnx, mxy - mny))
        except:
            return Rect(0, 0, 0, 0)
    
    def ambit(self, th=False, tv=False):
        """Get the calculated rect boundary of the DraftingPen;
        `th` means `(t)rue (h)orizontal`;
        `ty` means `(t)rue (v)ertical`;
        passing either ignores a non-bounds-derived frame
        in either dimension"""
        if self._frame:
            if (th or tv) and len(self.value) > 0:
                f = self._frame
                b = self.bounds()
                if th and tv:
                    return b
                elif th:
                    return Rect(b.x, f.y, b.w, f.h)
                else:
                    return Rect(f.x, b.y, f.w, b.h)
            else:
                return self._frame
        else:
            return self.bounds()
    
    def addFrame(self, frame, typographic=False, passthru=False):
        """Add a new frame to the DATPen, replacing any old frame. Passthru ignored, there for compatibility"""
        self._frame = frame
        if typographic:
            self.typographic = True
        return self
    
    def unended(self):
        if len(self.value) == 0:
            return True
        elif self.value[-1][0] not in ["endPath", "closePath"]:
            return True
        return False
    
    def fully_close_path(self):
        if self.value[-1][0] == "closePath":        
            start = self.value[0][-1][-1]
            end = self.value[-2][-1][-1]

            if start != end:
                self.value = self.value[:-1]
                self.lineTo(start)
                self.closePath()
        return self
    
    def reverse(self):
        """Reverse the winding direction of the pen."""
        if self.unended():
            self.closePath()
        dp = type(self)()
        rp = ReverseContourPen(dp)
        self.replay(rp)
        self.value = dp.value
        return self
    
    def __invert__(self):
        return self.reverse()
    
    def sh(self, s, subs={}):
        try:
            start = self.value[0][1][-1]
        except:
            start = None
        #print("SH", s, self.defs)
        res = sh(s, self, subs={"¬":self._last, "§":start, **subs})
        if res[0] == "∫":
            res = [self.single_pen_class().gs(res[1:])]
        return res
    
    def gss(self, s):
        dps = self.multi_pen_class()
        sh(s, ctx=self, dps=dps)
        for p in dps._pens:
            self.record(p)
        return self
    
    def gs(self, s, fn=None, tag=None, writer=None, ctx=None, dps=None, macros={}):
        ctx = ctx or self
        macros = {**self.macros, **macros}

        def expand_multisuffix(m):
            out = []
            arrows = list(m.group(2))
            for a in arrows:
                out.append(m.group(1)+a)
            return " ".join(out)

        def sp(_s):
            return [x.strip() for x in re.split(r"\s|\n", _s)]

        if isinstance(s, str):
            s = s
            s = re.sub(r"([\$\&]{1}[a-z]+)([↖↑↗→↘↓↙←•⍺⍵µ]{2,})", expand_multisuffix, s)
            #e = sh(s, ctx, dps)
            moves = sp(s)
        else:
            e = s
            moves = e
        
        def one_move(_e, move="lineTo"):
            if _e is None:
                return
            elif isinstance(_e, Point):
                getattr(self, move)(_e)
            elif isinstance(_e, Rect):
                self.rect(_e)
            elif isinstance(_e, Curve):
                _, b, c, d = _e
                self.curveTo(b, c, d)
            elif isinstance(_e, str):
                getattr(self, _e)()
            elif _e[0][0] == "∑":
                    macro = "".join(_e[0][1:])
                    if macro in macros:
                        macro_fn = macros[macro]
                        macro_fn(self, *_e[1:])
                    else:
                        raise Exception("unrecognized macro '" + macro + "'")
            elif len(_e) >= 3:
                if len(_e) >= 5:
                    self.interpCurveTo(*_e)
                else:
                    self.boxCurveTo(*_e)

        locals = {}
        mvs = [moves[0]]
        if isinstance(mvs[0], str):
            res = sh(mvs[0], ctx, dps)
        else:
            res = [mvs[0]]
        one_move(res[0], move="moveTo")

        try:
            start = self.value[0][1][-1]
        except:
            start = None

        for _m in moves[1:]:
            last = self._last
            ctx._last = last
            if isinstance(_m, str):
                res = sh(_m, ctx, dps, subs={"¬":last,"§":start})
            else:
                res = [_m]
            if res:
                one_move(res[0], move="lineTo")
        
        if self.unended():
            self.closePath()

        if tag:
            self.tag(tag)
        if fn:
            fn(self)
        return self

    def moveTo(self, p0):
        super().moveTo(p0)
        self._last = p0
        return self

    def lineTo(self, p1):
        if len(self.value) == 0:
            super().moveTo(p1)
        else:
            super().lineTo(p1)
        self._last = p1
        return self

    def qCurveTo(self, *points):
        super().qCurveTo(*points)
        self._last = points[-1]
        return self

    def curveTo(self, *points):
        super().curveTo(*points)
        self._last = points[-1]
        return self

    def closePath(self):
        super().closePath()
        return self

    def endPath(self):
        super().endPath()
        return self

    def addComponent(self, glyphName, transformation):
        if hasattr(self, "_glyphSet") and self._glyphSet:
            if glyphName in self._glyphSet:
                dp = DraftingPen().glyph(self._glyphSet[glyphName], self._glyphSet).transform(transformation)
                self.record(dp)
            #super().addComponent(glyphName, transformation)
        else:
            print("no glyphset, cannot add component")
        return self

    def replay(self, pen):
        super().replay(pen)
        return self
    
    def record(self, pen):
        """Play a pen into this pen, meaning that pen will be added to this one’s value."""
        if hasattr(pen, "_pens"):
            for p in pen:
                self.record(p)
        elif pen:
            pen.replay(self)
        return self
    
    def glyph(self, glyph, glyphSet=None):
        """Play a glyph (like from `defcon`) into this pen."""
        if glyphSet:
            self._glyphSet = glyphSet
        glyph.draw(self)
        return self
    
    def to_glyph(self, name=None, width=None, allow_blank=False):
        """
        Create a glyph (like from `defcon`) using this pen’s value.
        *Warning*: if path is unended, closedPath will be called
        """
        from defcon import Glyph
        if not allow_blank:
            if self.unended():
                self.closePath()
        bounds = self.bounds()
        glyph = Glyph()
        glyph.name = name
        glyph.width = width or bounds.w
        try:
            sp = glyph.getPen()
            self.replay(sp)
        except AssertionError:
            if not allow_blank:
                print(">>>blank glyph:", glyph.name)
        return glyph
    
    def rect(self, rect):
        """Rectangle primitive — `moveTo/lineTo/lineTo/lineTo/closePath`"""
        self.moveTo(rect.point("SW").xy())
        self.lineTo(rect.point("SE").xy())
        self.lineTo(rect.point("NE").xy())
        self.lineTo(rect.point("NW").xy())
        self.closePath()
        return self
    
    def roundedRect(self, rect, hr, vr=None):
        """Rounded rectangle primitive"""
        if vr is None:
            vr = hr
        l, b, w, h = Rect(rect)
        r, t = l + w, b + h
        K = 4 * (math.sqrt(2)-1) / 3
        circle = hr == 0.5 and vr == 0.5
        if hr <= 0.5:
            hr = w * hr
        if vr <= 0.5:
            vr = h * vr
        self.moveTo((l + hr, b))
        if not circle:
            self.lineTo((r - hr, b))
        self.curveTo((r+hr*(K-1), b), (r, b+vr*(1-K)), (r, b+vr))
        if not circle:
            self.lineTo((r, t-vr))
        self.curveTo((r, t-vr*(1-K)), (r-hr*(1-K), t), (r-hr, t))
        if not circle:
            self.lineTo((l+hr, t))
        self.curveTo((l+hr*(1-K), t), (l, t-vr*(1-K)), (l, t-vr))
        if not circle:
            self.lineTo((l, b+vr))
        self.curveTo((l, b+vr*(1-K)), (l+hr*(1-K), b), (l+hr, b))
        self.closePath()
        return self
    
    def oval(self, rect):
        """Oval primitive"""
        self.roundedRect(rect, 0.5, 0.5)
        return self

    def line(self, points, moveTo=True, endPath=True):
        """Syntactic sugar for `moveTo`+`lineTo`(...)+`endPath`; can have any number of points"""
        if isinstance(points, Line):
            points = list(points)
        if len(points) == 0:
            return self
        if len(self.value) == 0 or moveTo:
            self.moveTo(points[0])
        else:
            self.lineTo(points[0])
        for p in points[1:]:
            self.lineTo(p)
        if endPath:
            self.endPath()
        return self
    
    def hull(self, points):
        """Same as `DraftingPen.line` but calls closePath instead of endPath`"""
        self.moveTo(points[0])
        for pt in points[1:]:
            self.lineTo(pt)
        self.closePath()
        return self

    def interpCurveTo(self, p1, f1, p2, f2, to, inset=0):
        a = Point(self.value[-1][-1][-1])
        d = Point(to)
        pl = Line(p1, p2).inset(inset)
        b = Line(a, pl.start).t(f1/100)
        c = Line(d, pl.end).t(f2/100)
        return self.curveTo(b, c, d)
    
    def boxCurveTo(self, point, factor, pt, po=(0, 0), mods={}, flatten=False):
        if flatten:
            self.lineTo(pt)
            return self
        
        a = Point(self.value[-1][-1][-1])
        d = Point(pt)
        box = Rect.FromMnMnMxMx([min(a.x, d.x), min(a.y, d.y), max(a.x, d.x), max(a.y, d.y)])
        try:
            f1, f2 = factor
        except TypeError:
            if isinstance(factor, Atom):
                f1, f2 = (factor[0], factor[0])
            else:
                f1, f2 = (factor, factor)

        if isinstance(point, str):
            p = box.point(point)
            p1, p2 = (p, p)
        elif isinstance(point, Point):
            p1, p2 = point, point
        else:
            p1, p2 = point
            p1 = box.point(p1)
            p2 = box.point(p2)
        
        p1 = p1.offset(*po)
        p2 = p2.offset(*po)
        
        b = a.interp(f1, p1)
        c = d.interp(f2, p2)

        mb = mods.get("b")
        mc = mods.get("c")
        if mb:
            b = mb(b)
        elif mc:
            c = mc(c)
        
        self.curveTo(b, c, d)
        return self
    
    def vl(self, value):
        self.value = value
        return self
    
    def replace_with(self, other):
        return self.vl(other.value)
    
    def pvl(self):
        for idx, (_, pts) in enumerate(self.value):
            if len(pts) > 0:
                self.value[idx] = list(self.value[idx])
                self.value[idx][-1] = [Point(p) for p in self.value[idx][-1]]
        return self
    
    def add_data(self, key, value=None):
        if value is None:
            self.data = key
        else:
            self.data[key] = value
        return self
    
    def copy(self, with_data=False):
        dp = self.single_pen_class()
        self.replay(dp)
        for tag, attrs in self.attrs.items():
            dp.attr(tag, **attrs)
        
        dp.glyphName = self.glyphName
        dp.defs = self.defs

        if with_data:
            dp.data = self.data
            if self._frame:
                dp._frame = self._frame
            if hasattr(self, "macros"):
                dp.macros = self.macros
            if self.typographic:
                dp.typographic = True
        return dp
    
    def cast(self, _class, *args):
        """Quickly cast to a (different) subclass."""
        #if hasattr(self, "_pens"):
        #    return _class(self._pens)
        res = _class(self, *args)
        res.attrs = deepcopy(self.attrs)
        return res
    
    def pen(self):
        """Return a single-pen representation of this pen(set)."""
        return self
    
    def to_pen(self):
        return self.pen()
    
    def round(self):
        """Round the values of this pen to integer values."""
        return self.round_to(1)

    def round_to(self, rounding):
        """Round the values of this pen to nearest multiple of rounding."""
        def rt(v, mult):
            rndd = float(round(v / mult) * mult)
            if rndd.is_integer():
                return int(rndd)
            else:
                return rndd
        
        rounded = []
        for t, pts in self.value:
            _rounded = []
            for p in pts:
                if p:
                    x, y = p
                    _rounded.append((rt(x, rounding), rt(y, rounding)))
                else:
                    _rounded.append(p)
            rounded.append((t, _rounded))
        
        self.value = rounded
        return self

    def transform(self, transform, transformFrame=True):
        """Perform an arbitrary transformation on the pen, using the fontTools `Transform` class."""
        op = RecordingPen()
        tp = TransformPen(op, transform)
        self.replay(tp)
        self.value = op.value
        if transformFrame and self._frame:
            self._frame = self._frame.transform(transform)
        img = self.img()
        if img:
            img["rect"] = img["rect"].transform(transform)
        return self
    
    def align(self, rect, x="mdx", y="mdy", th=True, tv=False, transformFrame=True, h=None, return_offset=False):
        r = self.ambit(th, tv)
        if h is not None:
            r = r.seth(h)
        offset = align(r, rect, x, y)
        self.translate(*offset, transformFrame=transformFrame)
        if return_offset:
            return offset
        return self
    
    def x_align_to_frame(self, x=Edge.CenterX, th=0):
        if self._frame:
            return self.align(self.ambit(th=th, tv=0), x=x, transformFrame=1, th=1)
        else:
            raise Exception("No Frame")

    # deprecated camelcase
    xAlignToFrame = x_align_to_frame
    
    def translate(self, x, y=None, transformFrame=True):
        """Translate this shape by `x` and `y` (pixel values)."""
        if y is None:
            y = x
        return self.transform(Transform(1, 0, 0, 1, x, y), transformFrame=transformFrame)
    
    offset = translate

    def offset_x(self, x):
        return self.translate(x, 0)
    
    def offset_y(self, y):
        return self.translate(0, y)
    
    def zero_translate(self, th=1, tv=0):
        x, y, _, _ = self.ambit(th=th, tv=tv)
        self.translate(-x, -y)
        return self
    
    def center_on_point(self, rect, pt, interp=1):
        return self.translate(norm(interp, 0, rect.w/2-pt[0]), norm(interp, 0, rect.h/2-pt[1]))
    
    def skew(self, x=0, y=0, point=None):
        t = Transform()
        if not point:
            point = self.bounds().point("C") # maybe should be getFrame()?
        t = t.translate(point.x, point.y)
        t = t.skew(x, y)
        t = t.translate(-point.x, -point.y)
        return self.transform(t)
    
    def rotate(self, degrees, point=None):
        """Rotate this shape by a degree (in 360-scale, counterclockwise)."""
        t = Transform()
        if not point:
            point = self.bounds().point("C") # maybe should be getFrame()?
        elif isinstance(point, str):
            point = self.bounds().point(point)
        t = t.translate(point.x, point.y)
        t = t.rotate(math.radians(degrees))
        t = t.translate(-point.x, -point.y)
        return self.transform(t, transformFrame=False)
    
    def scale(self, scaleX, scaleY=None, point=None):
        """Scale this shape by a percentage amount (1-scale)."""
        t = Transform()
        if point != False:
            point = self.bounds().point("C") if point == None else point # maybe should be getFrame()?
            t = t.translate(point.x, point.y)
        t = t.scale(scaleX, scaleY or scaleX)
        if point != False:
            t = t.translate(-point.x, -point.y)
        return self.transform(t)
    
    def scaleToRect(self, rect, preserveAspect=True, shrink_only=False):
        """Scale this shape into a `Rect`."""
        bounds = self.bounds()
        h = rect.w / bounds.w
        v = rect.h / bounds.h
        if preserveAspect:
            scale = h if h < v else v
            if shrink_only and scale >= 1:
                return self
            return self.scale(scale)
        else:
            if shrink_only and (h >= 1 or v >= 1):
                return self
            return self.scale(h, v)
    
    def scaleToWidth(self, w, shrink_only=False):
        """Scale this shape horizontally"""
        b = self.bounds()
        if shrink_only and b.w < w:
            return self
        else:
            return self.scale(w / self.bounds().w, 1)
    
    def scaleToHeight(self, h, shrink_only=False):
        """Scale this shape horizontally"""
        b = self.bounds()
        if shrink_only and b.h < h:
            return self
        return self.scale(1, h / self.bounds().h)
    
    # PEN-BASED MODIFICATIONS
    
    def flatten(self, length=10):
        """
        Runs a fontTools `FlattenPen` on this pen
        """
        if length == 0:
            return self
        dp = type(self)()
        fp = FlattenPen(dp, approximateSegmentLength=length, segmentLines=True)
        self.replay(fp)
        self.value = dp.value
        return self
    
    def smooth(self, length=100):
        rp = RecordingPen()
        fp = SmoothPointsPen(rp)
        self.replay(fp)
        self.value = rp.value
        return self
    
    def explode(self):
        """Read each contour into its own DATPen; returns a DATPens"""
        dp = RecordingPen()
        ep = ExplodingPen(dp)
        self.replay(ep)
        dps = self.multi_pen_class()
        for p in ep._pens:
            dp = type(self)()
            dp.value = p
            dp.attrs = deepcopy(self.attrs)
            dps.append(dp)
        return dps
    
    def repeat(self, times=1):
        copy = self.copy()
        copy_0_move, copy_0_data = copy.value[0]
        copy.value[0] = ("lineTo", copy_0_data)
        self.value = self.value[:-1] + copy.value
        if times > 1:
            self.repeat(times-1)
        return self
    
    # Iteration-manipulation
    
    def take(self, slice):
        self.value = self.value[slice]
        return self
    
    def take_curve(self, idx):
        a = self.value[idx-1][-1][-1]
        b, c, d = self.value[idx][-1]
        return Curve(a, b, c, d)

    def ups(self):
        "Convert this single pen into a collection of pens, with one-pen in the collection (this pen)"
        dps = self.multi_pen_class()
        dps.append(self.copy())
        return dps
    
    def pens(self):
        """Return a set representation of this"""
        if hasattr(self, "_pens"):
            return self
        else:
            return self.ups()
    
    def collapse(self, levels=100, onself=False):
        return self.multi_pen_class([self])
    
    def mod_pt(self, vidx, pidx, fn):
        pt = Point(self.value[vidx][-1][pidx])
        if callable(fn):
            res = fn(pt)
        else:
            res = pt.offset(*fn)
        try:
            self.value[vidx][-1][pidx] = res
        except TypeError:
            self.pvl()
            self.value[vidx][-1][pidx] = res
        return self
    
    def mod_pts(self, rect, fn):
        self.map_points(fn, lambda p: p.inside(rect))
        return self
    
    # Contour manipulation

    def mod_contour(self, contour_index, mod_fn):
        exploded = self.explode()
        mod_fn(exploded[contour_index])
        self.value = exploded.implode().value
        return self
    
    def filter_contours(self, filter_fn):
        exploded = self.explode()
        keep = []
        for idx, c in enumerate(exploded):
            if filter_fn(idx, c):
                keep.append(c)
        self.value = self.multi_pen_class(keep).implode().value
        return self
    
    def slicec(self, contour_slice):
        self.value = self.multi_pen_class(self.explode()[contour_slice]).implode().value
        return self
    
    # Iterating

    def map(self, fn:Callable[[int, str, list], Tuple[str, list]]):
        for idx, (mv, pts) in enumerate(self.value):
            self.value[idx] = fn(idx, mv, pts)
        return self
    
    def filter(self, fn:Callable[[int, str, list], bool]):
        vs = []
        for idx, (mv, pts) in enumerate(self.value):
            if fn(idx, mv, pts):
                vs.append((mv, pts))
        self.value = vs
        return self
    
    def map_points(self, fn, filter_fn=None):
        idx = 0
        for cidx, c in enumerate(self.value):
            move, pts = c
            pts = list(pts)
            for pidx, p in enumerate(pts):
                x, y = p
                if filter_fn and not filter_fn(*p):
                    continue
                result = fn(idx, x, y)
                if result:
                    pts[pidx] = result
                idx += 1
            self.value[cidx] = (move, pts)
        return self
    
    def walk(self,
        callback:Callable[["DraftingPen", int, dict], None],
        depth=0,
        visible_only=False,
        parent=None,
        alpha=1
        ):
        if visible_only and not self._visible:
            return
        
        if parent:
            self._parent = parent
        
        alpha = self._alpha * alpha
        
        is_dps = hasattr(self, "_pens")
        if is_dps:
            callback(self, -1, dict(depth=depth, alpha=alpha))
            for pen in self._pens:
                pen.walk(callback, depth=depth+1, visible_only=visible_only, parent=self, alpha=alpha)
            callback(self, 1, dict(depth=depth, alpha=alpha))
        else:
            callback(self, 0, dict(depth=depth, alpha=alpha))
    
    def remove_blanks(self):
        print("REMOVE BLANKS PEN", self)
        """If this is blank, `return True` (for recursive calls from DATPens)."""
        return len(self.value) == 0
    
    def interpolate(self, value, other):
        if len(self.value) != len(other.value):
            raise Exception("Cannot interpolate / diff lens")
        vl = []
        for idx, (mv, pts) in enumerate(self.value):
            ipts = []
            for jdx, p in enumerate(pts):
                pta = Point(p)
                try:
                    ptb = Point(other.value[idx][-1][jdx])
                except IndexError:
                    print(">>>>>>>>>>>>> Can’t interpolate", idx, mv, "///", other.value[idx])
                    raise IndexError
                ipt = pta.interp(value, ptb)
                ipts.append(ipt)
            vl.append((mv, ipts))
        return type(self)().vl(vl)
    
    def Interpolate(instances, value):
        spread = len(instances)-1
        start = math.floor(value*spread)
        end = math.ceil(value*spread)
        v = value*spread-start
        return instances[start].interpolate(v, instances[end])
    
    # GEOMETRICAL

    def nsew(self):
        pts = [el[1][-1] for el in self.value if len(el[1]) > 0]
        
        lines = []
        for i, p in enumerate(pts):
            if i + 1 == len(pts):
                lines.append(Line(p, pts[0]))
            else:
                lines.append(Line(p, pts[i+1]))
        
        mnx, mny, mxx, mxy = self.bounds().mnmnmxmx()
        min_ang = min([l.ang for l in lines])
        max_ang = max([l.ang for l in lines])
        #for idx, l in enumerate(lines):
        #    print(idx, ">", l.ang, min_ang, max_ang)
        xs = [l for l in lines if l.ang < 0.25 or l.ang > 2.5]
        ys = [l for l in lines if 1 < l.ang < 2]

        if len(ys) == 2 and len(xs) < 2:
            xs = [l for l in lines if l not in ys]
        elif len(ys) < 2 and len(xs) == 2:
            ys = [l for l in lines if l not in xs]
        
        #for l in ys:
        #    print(l.ang)

        #print(len(xs), len(ys))
        #print("--------------------")

        n = [l for l in xs if l.start.y == mxy or l.end.y == mxy][0]
        s = [l for l in xs if l.start.y == mny or l.end.y == mny][0]
        e = [l for l in ys if l.start.x == mxx or l.end.x == mxx][0]
        w = [l for l in ys if l.start.x == mnx or l.end.x == mnx][0]
        return n, s, e, w
    
    def avg(self):
        self.pvl()
        pts = []
        for _, _pts in self.value:
            if len(_pts) > 0:
                pts.extend(_pts)
        n = len(pts)
        #print("AVG", self, self.value)
        return Point(
            sum([p.x for p in pts])/n,
            sum([p.y for p in pts])/n)

    def point(self, pt):
        n, s, e, w = self.nsew()
        if pt == "NE":
            return n.pe
        elif pt == "NW":
            return n.pw
        elif pt == "SE":
            return s.pe
        elif pt == "SW":
            return s.pw
        elif pt == "N":
            return n.mid
        elif pt == "S":
            return s.mid
        elif pt == "E":
            return e.mid
        elif pt == "W":
            return w.mid

    @property
    def pne(self): return self.point("NE")
    @property
    def pnw(self): return self.point("NW")
    @property
    def psw(self): return self.point("SW")
    @property
    def pse(self): return self.point("SE")
    @property
    def pn(self): return self.point("N")
    @property
    def ps(self): return self.point("S")
    @property
    def pe(self): return self.point("E")
    @property
    def pw(self): return self.point("W")
    @property
    def en(self): return self.nsew()[0]
    @property
    def es(self): return self.nsew()[1]
    @property
    def ee(self): return self.nsew()[2]
    @property
    def ew(self): return self.nsew()[3]
    
    @property
    def ecx(self):
        n, s, e, w = self.nsew()
        return e.interp(0.5, w.reverse())
    
    @property
    def ecy(self):
        n, s, e, w = self.nsew()
        return n.interp(0.5, s.reverse())
    
    def edge(self, e):
        e = e.lower()
        if e == "n":
            return self.en
        elif e == "s":
            return self.es
        elif e == "e":
            return self.ee
        elif e == "w":
            return self.ew
    
    def shprop(self, s):
        if s in SH_UNARY_SUFFIX_PROPS:
            return SH_UNARY_SUFFIX_PROPS[s]
        return s
    
    def pinch(self, edge, inset):
        if isinstance(edge, str):
            e = getattr(self, self.shprop(edge))
        elif isinstance(edge, int):
            if edge == 0:
                e = self.en
            elif edge == 1:
                e = self.ee
            elif edge == 2:
                e = self.es
            elif edge == 3:
                e = self.ew
        ei = e.inset(inset)
        self.pvl()
        for idx, (mv, pts) in enumerate(self.value):
            for jdx, pt in enumerate(pts):
                if pt == e.start:
                    self.value[idx][1][jdx] = ei.start
                elif pt == e.end:
                    self.value[idx][1][jdx] = ei.end
        return self

    # COMPUTATIONAL OBJECT
    
    def define(self, *args, **kwargs):
        return self.context_record("$", "defs", None, *args, **kwargs)
    
    def declare(self, *whatever):
        # TODO do something with what's declared somehow?
        return self
    
    def macro(self, **kwargs):
        for k, v in kwargs.items():
            self.macros[k] = v
        return self

    def guide(self, grid:Grid):
        for k, v in grid.keyed.items():
            setattr(self, k, v)
        return self
    
    def print(self, *args):
        for a in args:
            if callable(a):
                print(a(self))
            else:
                print(a)
        return self

    def noop(self, *args, **kwargs):
        """Does nothing"""
        return self
    
    def sleep(self, time):
        """Sleep call within the chain (if you want to measure something)"""
        sleep(time)
        return self

    def chain(self,
        fn:Callable[["DraftingPen"], None],
        *args
        ):
        """
        For simple take-one callback functions in a chain
        """
        if fn:
            try:
                fn, metadata = fn
            except TypeError:
                metadata = {}
            
            res = fn(self, *args)
            if "returns" in metadata:
                return res
            elif isinstance(res, DraftingPen):
                return res
        return self
    
    ch = chain
    
    def cond(self, condition, if_true:Callable[["DraftingPen"], None], if_false:Callable[["DraftingPen"], None]=None):
        if callable(condition):
            condition = condition(self)
        # TODO make if_false optional
        if condition:
            if callable(if_true):
                if_true(self)
            else:
                self.gs(if_true)
        else:
            if if_false is not None:
                if callable(if_false):
                    if_false(self)
                else:
                    self.gs(if_false)
        return self

    # BOOLEAN OPERATIONS

    def _pathop(self, otherPen=None, operation=BooleanOp.XOR):
        self.value = calculate_pathop(self, otherPen, operation)
        return self
    
    def difference(self, otherPen):
        """Calculate and return the difference of this shape and another."""
        return self._pathop(otherPen=otherPen, operation=BooleanOp.Difference)
    
    def union(self, otherPen):
        """Calculate and return the union of this shape and another."""
        return self._pathop(otherPen=otherPen, operation=BooleanOp.Union)
    
    def xor(self, otherPen):
        """Calculate and return the XOR of this shape and another."""
        return self._pathop(otherPen=otherPen, operation=BooleanOp.XOR)
    
    def reverseDifference(self, otherPen):
        """Calculate and return the reverseDifference of this shape and another."""
        return self._pathop(otherPen=otherPen, operation=BooleanOp.ReverseDifference)
    
    def intersection(self, otherPen):
        """Calculate and return the intersection of this shape and another."""
        return self._pathop(otherPen=otherPen, operation=BooleanOp.Intersection)
    
    def removeOverlap(self):
        """Remove overlaps within this shape and return itself."""
        return self._pathop(otherPen=None, operation=BooleanOp.Simplify)
    
    # ATTRIBUTES

    def clearAttrs(self):
        """Remove all styling."""
        self.attrs = OrderedDict()
        self.attr("default", fill=(1, 0, 0.5))
        return self
    
    def allStyledAttrs(self, style=None):
        if style and style in self.attrs:
            attrs = self.attrs[style]
        else:
            attrs = self.attrs["default"]
        return attrs

    def attr(self, tag=None, field=None, **kwargs):
        """Set a style attribute on the pen."""
        if not tag:
            if hasattr(self, "_current_attr_tag"): # TODO temporary for pickled pens
                tag = self._current_attr_tag
            else:
                tag = "default"

        if field: # getting, not setting
            return self.attrs.get(tag).get(field)
        
        attrs = dict(shadow=None)
        if tag and self.attrs.get(tag):
            attrs = self.attrs[tag]
        else:
            self.attrs[tag] = attrs
        for k, v in kwargs.items():
            if v:
                if k == "fill":
                    attrs[k] = normalize_color(v)
                elif k == "stroke":
                    existing = attrs.get("stroke", {})
                    if not isinstance(v, dict):
                        c = normalize_color(v)
                        if hasattr(c, "a"):
                            default_w = 1 if c.a > 0 else 0
                        else:
                            default_w = 1
                        attrs[k] = dict(color=c, weight=existing.get("weight", default_w))
                    else:
                        c = normalize_color(v.get("color", 0))
                        if hasattr(c, "a"):
                            default_w = 1 if c.a > 0 else 0
                        else:
                            default_w = 1
                        attrs[k] = dict(weight=v.get("weight", existing.get("weight",default_w)), color=c)
                elif k == "strokeWidth":
                    if "stroke" in attrs:
                        attrs["stroke"]["weight"] = v
                        #if attrs["stroke"]["color"].a == 0:
                        #    attrs["stroke"]["color"] = normalize_color((1, 0, 0.5))
                    else:
                        attrs["stroke"] = dict(color=normalize_color((1, 0, 0.5)), weight=v)
                elif k == "shadow":
                    if "color" in v:
                        v["color"] = normalize_color(v["color"])
                    attrs[k] = v
                else:
                    attrs[k] = v
        return self
    
    def lattr(self, tag, fn: Callable[["DraftingPen"], Optional["DraftingPen"]]):
        was_tag = self._current_attr_tag
        self._current_attr_tag = tag
        fn(self)
        self._current_attr_tag = was_tag
        return self
    
    # def calc_alpha(self):
    #     a = self._alpha
    #     p = self._parent
    #     while p:
    #         a = a * p._alpha
    #         p = p._parent
    #     return a
    
    def v(self, v):
        self.visible(bool(v))
        return self
    
    def a(self, v):
        self._alpha = v
        return self

    def f(self, *value):
        """Get/set a (f)ill"""
        if value:
            return self.attr(fill=value)
        else:
            return self.attr(field="fill")
    
    fill = f
    
    def s(self, *value):
        """Get/set a (s)troke"""
        if value:
            return self.attr(stroke=value)
        else:
            return self.attr(field="stroke")
    
    stroke = s
    
    def sw(self, value):
        """Get/set a (s)troke (w)idth"""
        if value is not None:
            return self.attr(strokeWidth=value)
        else:
            return self.attr(field="strokeWidth")
    
    strokeWidth = sw

    def img(self, src=None, rect=Rect(0, 0, 500, 500), pattern=True, opacity=1.0):
        """Get/set an image fill"""
        if src:
            return self.attr(image=dict(src=src, rect=rect, pattern=pattern, opacity=opacity))
        else:
            return self.attr(field="image")
    
    def img_opacity(self, opacity, key="default"):
        img = self.attr(key, "image")
        if not img:
            raise Exception("No image found")
        self.attrs[key]["image"]["opacity"] = opacity
        return self
    
    image = img

    def shadow(self, radius=10, color=(0, 0.3), clip=None):
        return self.attr(shadow=dict(color=normalize_color(color), radius=radius, clip=clip))
    
    def all_pens(self):
        pens = []
        if hasattr(self, "_pens"):
            pens = self.collapse()._pens
        if isinstance(self, self.single_pen_class):
            pens = [self]
        
        for pen in pens:
            if pen:
                if hasattr(pen, "_pens"):
                    for _p in pen.collapse()._pens:
                        if _p:
                            yield _p
                else:
                    yield pen
    
    # Fun pen manipulations

    def outline(self, offset=1, drawInner=True, drawOuter=True, cap="square", miterLimit=None):
        """AKA expandStroke"""
        op = OutlinePen(None, offset=offset, optimizeCurve=True, cap=cap, miterLimit=miterLimit)
        self.replay(op)
        op.drawSettings(drawInner=drawInner, drawOuter=drawOuter)
        g = op.getGlyph()
        p = self.single_pen_class()
        g.draw(p)
        self.value = p.value
        return self
    
    ol = outline
    
    def project(self, angle, width):
        offset = polarCoord((0, 0), math.radians(angle), width)
        self.translate(offset[0], offset[1])
        return self

    def castshadow(self, angle=-45, width=100, ro=1, fill=1):
        out = self.single_pen_class()
        tp = TranslationPen(out, frontAngle=angle, frontWidth=width)
        self.replay(tp)
        if fill:
            out.record(self.copy().project(angle, width))
        if ro:
            out.removeOverlap()
        self.value = out.value
        return self

    def grow(self, outline=10):
        out = self.copy().outline(outline)
        return self.record(out.reverse())
    
    def gridlines(self, rect, x=20, y=None, absolute=False):
        """Construct a grid in the pen using `x` and (optionally) `y` subdivisions"""
        xarg = x
        yarg = y or x
        if absolute:
            x = int(rect.w / xarg)
            y = int(rect.h / yarg)
        else:
            x = xarg
            y = yarg
        
        for _x in rect.subdivide(x, "minx"):
            if _x.x > 0 and _x.x > rect.x:
                self.line([_x.point("NW"), _x.point("SW")])
        for _y in rect.subdivide(y, "miny"):
            if _y.y > 0 and _y.y > rect.y:
                self.line([_y.point("SW"), _y.point("SE")])
        return self.f(None).s(0, 0.1).sw(3)
    
    # Some curvy/bendy things

    def subsegment(self, start=0, end=1):
        """Return a subsegment of the pen based on `t` values `start` and `end`"""
        cc = CurveCutter(self)
        start = 0
        end = end * cc.calcCurveLength()
        pv = cc.subsegment(start, end)
        self.value = pv
        return self
    
    def point_t(self, t=0.5):
        """Get point value for time `t`"""
        cc = CurveCutter(self)
        start = 0
        tv = t * cc.calcCurveLength()
        p, tangent = cc.subsegmentPoint(start=0, end=tv)
        return p, tangent
    
    def split_t(self, t=0.5):
        a = self.value[0][-1][0]
        b, c, d = self.value[-1][-1]
        return splitCubicAtT(a, b, c, d, t)
    
    def add_pt_t(self, cuidx, t):
        cidx = 0
        insert_idx = -1
        c1, c2 = None, None

        for idx, (mv, pts) in enumerate(self.value):
            if mv == "curveTo":
                if cidx == cuidx:
                    insert_idx = idx
                    a = self.value[idx-1][-1][-1]
                    b, c, d = pts
                    c1, c2 = splitCubicAtT(a, b, c, d, t)
                cidx += 1
            elif mv == "lineTo":
                if cidx == cuidx:
                    insert_idx = idx
                    a = self.value[idx-1][-1][-1]
                    b = pts[0]
                    l = Line(a, b)
                    c1 = [l.t(0.5)]
                    c2 = [b]
                cidx += 1
        
        if c2:
            if len(c2) > 1:
                self.value[insert_idx] = ("curveTo", c1[1:])
                self.value.insert(insert_idx+1, ("curveTo", c2[1:]))
            else:
                self.value[insert_idx] = ("lineTo", c1)
                self.value.insert(insert_idx+1, ("lineTo", c2))
        return self
    
    def length(self, t=1):
        """Get the length of the curve for time `t`"""
        cc = CurveCutter(self)
        start = 0
        tv = t * cc.calcCurveLength()
        return tv
    
    def nonlinear_transform(self, fn):
        for idx, (move, pts) in enumerate(self.value):
            if len(pts) > 0:
                _pts = []
                for _pt in pts:
                    x, y = _pt
                    _pts.append(fn(x, y))
                self.value[idx] = (move, _pts)
        return self
    
    nlt = nonlinear_transform


import os
import pickle
import sys
from fontTools.designspaceLib import DesignSpaceDocument
from fontTools.fontBuilder import FontBuilder
from fontTools.ttLib import TTFont, newTable
from fontTools import varLib
from fontTools.varLib.errors import VarLibError


def compileDSToFont(dsPath, ttFolder):
    doc = DesignSpaceDocument.fromfile(dsPath)
    doc.findDefault()

    ufoPathToTTPath = getTTPaths(doc, ttFolder)

    for source in doc.sources:
        if source.layerName is None:
            ttPath = ufoPathToTTPath[source.path]
            if not os.path.exists(ttPath):
                raise FileNotFoundError(ttPath)
            source.font = TTFont(ttPath, lazy=False)

    assert doc.default.font is not None
    if "name" not in doc.default.font:
        doc.default.font["name"] = newTable("name")  # This is the template for the VF, and needs a name table

    if any(s.layerName is not None for s in doc.sources):
        fb = FontBuilder(unitsPerEm=doc.default.font["head"].unitsPerEm)
        fb.setupGlyphOrder(doc.default.font.getGlyphOrder())
        fb.setupPost()  # This makes sure we store the glyph names
        font = fb.font
        for source in doc.sources:
            if source.font is None:
                source.font = font

    try:
        ttFont, masterModel, _ = varLib.build(doc, exclude=['MVAR', 'HVAR', 'VVAR', 'STAT'])
    except VarLibError as e:
        if 'GSUB' in e.args:
            extraExclude = ['GSUB']
        elif 'GPOS' in e.args:
            extraExclude = ['GPOS', 'GDEF']
        else:
            raise
        print(f"{e!r}", file=sys.stderr)
        print(f"Error while building {extraExclude[0]} table, trying again without {' and '.join(extraExclude)}.",
              file=sys.stderr)
        ttFont, masterModel, _ = varLib.build(doc, exclude=['MVAR', 'HVAR', 'VVAR', 'STAT'] + extraExclude)

    # Our client needs the masterModel, so we save a pickle into the font
    ttFont["MPcl"] = newTable("MPcl")
    ttFont["MPcl"].data = pickle.dumps(masterModel)

    return ttFont


def compileDSToPath(dsPath, ttFolder, ttPath):
    ttFont = compileDSToFont(dsPath, ttFolder)
    ttFont.save(ttPath, reorderTables=False)


def getTTPaths(doc, ttFolder):
    ufoPaths = sorted({s.path for s in doc.sources if s.layerName is None})
    return {ufoPath: os.path.join(ttFolder, f"master_{index}.ttf")
            for index, ufoPath in enumerate(ufoPaths)}



from random import random
from coldtype.color.html import NAMED_COLORS
from coldtype.interpolation import norm

try:
    import skia
except ImportError:
    skia = None

try:
    from fontTools.ttLib.tables.C_P_A_L_ import Color as FTCPALColor
except ImportError:
    FTCPALColor = None


def norm(value, start, stop):
    return start + (stop-start) * value

def lerp(start, stop, amt):
    return float(amt-start) / float(stop-start)


# inspired by https://github.com/xav/Grapefruit/blob/master/grapefruit.py
# but more shorthand-oriented


def hue2rgb(n1, n2=None, h=None):
    h %= 6.0
    if h < 1.0:
        return n1 + ((n2-n1) * h)
    if h < 3.0:
        return n2
    if h < 4.0:
        return n1 + ((n2-n1) * (4.0 - h))
    return n1


def hsl_to_rgb(h, s=0, l=0):
    if s == 0:
        return (l, l, l)
    if l < 0.5:
        n2 = l * (1.0 + s)
    else:
        n2 = l+s - (l*s)
    n1 = (2.0 * l) - n2
    h /= 60.0
    r = hue2rgb(n1, n2, h + 2)
    g = hue2rgb(n1, n2, h)
    b = hue2rgb(n1, n2, h - 2)
    return (r, g, b)


def rgb_to_hsl(r, g=None, b=None):
    minVal = min(r, g, b)
    maxVal = max(r, g, b)

    l = (maxVal + minVal) / 2.0
    if minVal == maxVal:
        return (0.0, 0.0, l)

    d = maxVal - minVal

    if l < 0.5:
        s = d / (maxVal + minVal)
    else:
        s = d / (2.0 - maxVal - minVal)

    dr, dg, db = [(maxVal-val) / d for val in (r, g, b)]

    if r == maxVal:
        h = db - dg
    elif g == maxVal:
        h = 2.0 + dr - db
    else:
        h = 4.0 + dg - dr

    h = (h*60.0) % 360.0
    return (h, s, l)


class Color:
    def __init__(self, *values):
        r, g, b = [float(v) for v in values[:3]]
        self.r = float(values[0])
        self.g = float(values[1])
        self.b = float(values[2])
        if len(values) > 3:
            self.a = float(values[3])
        else:
            self.a = 1
        h, s, l = rgb_to_hsl(r, g, b)
        self.h = h
        self.s = s
        self.l = l
        self.html = self.to_html()
    
    def __eq__(self, other):
        if isinstance(other, Color):
            return self.r == other.r and self.g == other.g and self.b == other.b and self.a == other.a
        else:
            return False
    
    def __str__(self):
        return "<Color:rgb({:.1f},{:.1f},{:.1f})/a={:.1f}>".format(self.r, self.g, self.b, self.a)
    
    def to_code(self):
        if self.a == 1:
            if self.s == 0:
                return f"bw({self.l})"
        elif self.a < 1:
            if self.s == 0:
                return f"bw({self.l}, {self.a})"
        
        return f"hsl({round(self.h/360.0, 2)}, {round(self.s, 2)}, {round(self.l, 2)}, {round(self.a, 2)})"

    def with_alpha(self, alpha):
        return Color(self.r, self.g, self.b, alpha)

    def ints(self):
        return [self.r*255, self.g*255, self.b*255, self.a]
    
    def __getitem__(self, index):
        return [self.r, self.g, self.b, self.a][index]

    def from_rgb(r, g, b, a=1):
        return Color(r, g, b, a)

    def from_html(html, a=1):
        html = html.strip().lower()
        if html[0] == '#':
            html = html[1:]
        elif html in NAMED_COLORS:
            html = NAMED_COLORS[html][1:]
        if len(html) == 6:
            rgb = html[:2], html[2:4], html[4:]
        elif len(html) == 3:
            rgb = ['%c%c' % (v, v) for v in html]
        else:
            raise ValueError("input #%s is not in #RRGGBB format" % html)
        return Color(*[(int(n, 16) / 255.0) for n in rgb], a)
    
    def to_html(self):
        return '#%02x%02x%02x' % tuple((min(round(v*255), 255) for v in (self.r, self.g, self.b)))
    
    def lighter(self, level):
        return Color.from_hsl(self.h, self.s, min(self.l + level, 1), self.a)
    
    def desaturate(self, level):
        return Color.from_hsl(self.h, max(self.s - level, 0), self.l, self.a)
    
    def saturate(self, level):
        return Color.from_hsl(self.h, min(self.s + level, 1), self.l, self.a)
    
    def darker(self, level):
        return Color.from_hsl(self.h, self.s, max(self.l - level, 0), self.a)

    def from_hsl(h, s, l, a=1):
        r, g, b = hsl_to_rgb(h, s, l)
        return Color(r, g, b, a)
    
    def rgba(self):
        return self.r, self.g, self.b, self.a
    
    def hsl_interp(self, v, other):
        return hsl(norm(v, self.h, other.h)/360.0, norm(v, self.s, other.s), norm(v, self.l, other.l), norm(v, self.a, other.a))
    
    def rgb_interp(self, v, other):
        return rgb(norm(v, self.r, other.r), norm(v, self.g, other.g), norm(v, self.b, other.b), norm(v, self.a, other.a))
    
    def __repr__(self):
        return "<Color:({:0.2f},{:0.2f},{:0.2f})/({:0.2f},{:0.2f},{:0.2f})a={:0.2f}/>".format(self.r, self.g, self.b, self.h, self.s, self.l, self.a)
    
    def skia(self):
        if skia:
            return skia.Color4f(self.r, self.g, self.b, self.a)
        else:
            raise Exception("Skia installation not found")

def lighten_max(color, maxLightness=0.55):
    return Color.from_hsl(color.h, color.s, max(maxLightness, color.l))


def color_var(*rgba):
    c = [random() if x == -1 or x == "random" or x == "rand" or x == "R" else x for x in rgba]
    if len(c) == 1:
        return Color.from_rgb(c[0], c[0], c[0])
    elif len(c) == 2:
        return Color.from_rgb(c[0], c[0], c[0], c[1])
    elif len(c) == 3:
        return Color.from_rgb(c[0], c[1], c[2])
    elif len(c) == 4:
        return Color.from_rgb(c[0], c[1], c[2], c[3])


def hex_to_tuple(h):
    return tuple([c/255 for c in (palette.r, palette.g, palette.b, palette.a)])


def find_random(v):
    if isinstance(v, str):
        if v == "random" or v == "r":
            return random()
        elif v.startswith("r"):
            v = v[1:]
            if "-" in v:
                limits = [float(x.strip()) for x in v.split("-")]
                return random() * (limits[1]-limits[0]) + limits[0]
            elif "," in v:
                options = [float(x.strip()) for x in v.split(",")]
                return options[randint(0, len(options))]
    try:
        return float(v)
    except:
        return v


def normalize_color(v):
    if v is None:
        return Color.from_rgb(0,0,0,0)
    elif isinstance(v, Color):
        return v
    elif isinstance(v, Gradient):
        return v
    elif isinstance(v, float) or isinstance(v, int):
        return Color.from_rgb(v, v, v)
    elif FTCPALColor and isinstance(v, FTCPALColor):
        return Color.from_rgb(v.red/255, v.green/255, v.blue/255, v.alpha/255)
    elif isinstance(v, str):
        if v == "random" or v == -1:
            return Color.from_rgb(random(), random(), random())
        elif v == "none":
            return Color.from_rgb(0,0,0,0)
        else:
            return Color.from_html(v)
    else:
        if len(v) == 1:
            if v[0] == "random":
                return Color.from_rgb(random(), random(), random(), 1)
            if v[0] == None:
                return Color.from_rgb(0,0,0,0)
            elif isinstance(v[0], str):
                return Color.from_html(v[0])
            elif isinstance(v[0], Gradient):
                return v[0]
            else:
                try:
                    iter(v[0])
                    return normalize_color(v[0])
                except TypeError:
                    return Color.from_rgb(v[0], v[0], v[0])
        elif len(v) == 2:
            if v[0] == "random" or v[0] == -1:
                return Color.from_rgb(random(), random(), random(), float(v[1]))
            elif isinstance(v[0], str):
                return Color.from_html(v[0]).with_alpha(v[1])
            else:
                c = Color.from_rgb(v[0], v[0], v[0], v[1])
                return c
        else:
            if isinstance(v[0], complex):
                vs = [find_random(x) for x in v]
                return Color.from_hsl(v[0].imag*360, *vs[1:])
            if isinstance(v[0], str) and v[0].startswith("h"):
                v = list(v)
                v[0] = v[0][1:]
                vs = [find_random(x) for x in v]
                return Color.from_hsl(vs[0]*360, *vs[1:])
            else:
                vs = [random() if _v == "random" else _v for _v in v]
                return Color.from_rgb(*vs)


def hsl(h, s=0.5, l=0.5, a=1):
    return Color.from_hsl(h*360, s, l, a)

def rgb(r, g, b, a=1):
    return Color.from_rgb(r, g, b, a)

def bw(c, a=1):
    return Color.from_rgb(c, c, c, a)


class Gradient():
    def __init__(self, *stops):
        self.stops = []
        for c, p in stops:
            self.addStop(c, p)
    
    def addStop(self, color, point):
        self.stops.append([normalize_color(color), point])
    
    def Vertical(rect, a, b):
        return Gradient([a, rect.point("N")], [b, rect.point("S")])
    
    def Horizontal(rect, a, b):
        return Gradient([a, rect.point("W")], [b, rect.point("E")])
    
    def Random(rect, opacity=0.5):
        return Gradient([("random", opacity), rect.point("SE")], [("random", opacity), rect.point("NW")])
    
    V = Vertical
    H = Horizontal
    R = Random


from coldtype.pens.datpen import DATPen, DATPens

import coldtype.text.composer

from coldtype.geometry.rect import Rect
from coldtype.text.reader import Style, StyledString, SegmentedString, normalize_font_path, Font, FontNotFoundException
from coldtype.text.composer import Slug, Lockup, Graf, GrafStyle, T2L, Composer, StSt
#from coldtype.text.richtext import RichText


import srt, re
from timecode import Timecode
from datetime import datetime

from coldtype.time.sequence import Sequence, Clip, ClipTrack

def srt_to_frame(fps, st):
    tc = Timecode(fps,
        srt.timedelta_to_srt_timestamp(st).replace(",", "."))
    return tc.frame_number-86400-21600

class SRT(Sequence):
    def __init__(self, path, fps, storyboard=None, workarea_track=0):
        self.path = path
        subs = srt.parse(path.read_text())
        clips = []

        for sidx, s in enumerate(subs):
            clips.append(Clip(
                re.sub(r"<[^>]+>", "", s.content).strip(),
                srt_to_frame(fps, s.start),
                srt_to_frame(fps, s.end),
                idx=sidx,
                track=0))
        
        track = ClipTrack(self, clips, [])
        super().__init__(track.duration(), fps, storyboard, [track])


import unittest
from pathlib import Path
from coldtype.grid import Grid
from coldtype.geometry import *
from coldtype.text.composer import StSt
from coldtype.pens.svgpen import SVGPen

tf = Path(__file__).parent

class TestText(unittest.TestCase):
    def _test_glyph_names(self, font_path):
        ss = StSt("CDELOPTY", font_path, 100, wdth=0)
        ssps = ss#.pens()
        self.assertEqual(len(ssps), 8)
        self.assertEqual(ssps[0].glyphName, "C")
        self.assertEqual(ssps[-1].glyphName, "Y")

        fp = Path(font_path)
        op = (tf / f"ignorables/{fp.name}.svg")
        op.parent.mkdir(exist_ok=True)
        op.write_text(SVGPen.Composite(ssps, ssps.ambit(), viewBox=True))
        return ssps

    def test_format_equality(self):
        ttf = self._test_glyph_names("assets/ColdtypeObviously-VF.ttf")
        otf = self._test_glyph_names("assets/ColdtypeObviously_CompressedBlackItalic.otf")
        ufo = self._test_glyph_names("assets/ColdtypeObviously_CompressedBlackItalic.ufo")
        ds = self._test_glyph_names("assets/ColdtypeObviously.designspace")

        # TODO why isn't the ttf version equal to these?
        self.assertEqual(ufo[0].value, ds[0].value)
        self.assertEqual(ufo[-1].value, ds[-1].value)

if __name__ == "__main__":
    unittest.main()


try:
    import cv2
except ImportError:
    raise Exception("pip install opencv-python")

import numpy as np
from coldtype.img.skiaimage import skia, SkiaImage

def read_frame(cam):
    _, frame = cam.read()
    r_channel, g_channel, b_channel = cv2.split(frame)
    alpha_channel = np.ones(b_channel.shape, dtype=b_channel.dtype) * 255
    img_BGRA = cv2.merge((b_channel, g_channel, r_channel, alpha_channel))
    return SkiaImage(skia.Image.fromarray(img_BGRA))


try:
    import drawBot as db
except:
    pass

from coldtype.pens.drawbotpen import DrawBotPen

class RendererDrawBotPen(DrawBotPen):

    #def draw(self, scale=2, style=None):
    #    return super().draw(scale=scale, style=style)

    def Composite1(pens, rect, save_to, paginate=False, scale=2):
        db.newDrawing()
        rect = rect.scale(scale)
        if not paginate:
            db.newPage(rect.w, rect.h)
        for pen in RendererDrawBotPen.FindPens(pens):
            if paginate:
                db.newPage(rect.w, rect.h)
            RendererDrawBotPen(pen, rect).draw(scale=scale)
        db.saveImage(str(save_to))
        db.endDrawing()
    
    def Composite(pens, rect, save_to, scale=2):
        db.newDrawing()
        rect = rect.scale(scale)
        db.newPage(rect.w, rect.h)

        def draw(pen, state, data):
            if state == 0:
                RendererDrawBotPen(pen, rect).draw(scale=scale)
            elif state == -1:
                imgf = pen.data.get("imgf")
                if imgf:
                    im = db.ImageObject()
                    im.lockFocus()
                    db.size(rect.w+300, rect.h+300)
                    db.translate(150, 150)
                    db.scale(scale)
                    pen.data["im"] = im
            elif state == 1:
                imgf = pen.data.get("imgf")
                im = pen.data.get("im")
                if imgf and im:
                    im.unlockFocus()
                    imgf(im)
                    x, y = im.offset()
                    db.translate(-150, -150)
                    db.image(im, (x, y))
        
        if not hasattr(pens, "_pens"):
            pens = [pens]
        
        for dps in pens:
            dps.walk(draw)
        
        db.saveImage(str(save_to))
        db.endDrawing()


import weakref


_NotFoundToken = object()


class cachedProperty:

    """
        >>> class Foo:
        ...     @cachedProperty
        ...     def calcOnce(self):
        ...         print("calculating")
        ...         return 123
        ...
        >>> f = Foo()
        >>> f.calcOnce
        calculating
        123
        >>> f.calcOnce
        123
        >>> f.calcOnce = 134
        Traceback (most recent call last):
            ...
        AttributeError: calcOnce is read-only
        >>> del f.calcOnce
        >>> f.calcOnce
        calculating
        123
    """

    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.name, _NotFoundToken)
        if value is _NotFoundToken:
            value = self.func(obj)
            obj.__dict__[self.name] = value
        return value

    def __set__(self, obj, value):
        raise AttributeError(f"{self.name} is read-only")

    def __delete__(self, obj):
        # Mark cache as stale
        if self.name in obj.__dict__:
            del obj.__dict__[self.name]


class hookedProperty:

    """Property that calls a hook whenever its value changed, or
    when it got deleted.

    The `hook` should be a callable, that will be called with the
    owner of property as the first argument, and is therefore
    effectively a method of the owner.

    The hook will only be called when the value is actually
    different, so when clients set it to the same value several
    times in a row, the hook is only called once.

    Likewise, the hook itself can rely on the fact that the value
    has actually changed when it is called.
    """

    def __init__(self, hook, default=_NotFoundToken):
        self.hook = hook
        self.default = default

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, cls=None):
        try:
            return obj.__dict__[self.name]
        except KeyError:
            if self.default is _NotFoundToken:
                raise AttributeError(self.name)
            else:
                return self.default

    def __set__(self, obj, value):
        if self.name not in obj.__dict__ or obj.__dict__[self.name] != value:
            obj.__dict__[self.name] = value
            self.hook(obj)

    def __delete__(self, obj):
        try:
            del obj.__dict__[self.name]
        except KeyError:
            raise AttributeError(self.name)
        self.hook(obj)


class delegateProperty:

    def __init__(self, delegateAttributeName):
        self.delegateAttributeName = delegateAttributeName

    def __set_name__(self, owner, name):
        # This is Python >= 3.6
        self.propertyName = name

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        delegate = getattr(obj, self.delegateAttributeName)
        return getattr(delegate, self.propertyName)

    def __set__(self, obj, value):
        delegate = getattr(obj, self.delegateAttributeName)
        setattr(delegate, self.propertyName, value)

    def __delete__(self, obj):
        delegate = getattr(obj, self.delegateAttributeName)
        delattr(delegate, self.propertyName)


class weakrefCallbackProperty:

    def __init__(self, doc=None):
        self.__doc__ = doc

    def __set_name__(self, owner, name):
        self.weakrefCallbackName = "_weakrefCallback_" + name

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        weakrefCallback = getattr(obj, self.weakrefCallbackName, None)
        if weakrefCallback is None:
            return None
        return weakrefCallback()

    def __set__(self, obj, value):
        if value is None:
            self.__delete__(obj)
        else:
            try:
                weakMethod = weakref.WeakMethod(value)
            except TypeError:
                setattr(obj, self.weakrefCallbackName, lambda: value)
            else:
                setattr(obj, self.weakrefCallbackName, weakMethod)

    def __delete__(self, obj):
        try:
            delattr(obj, self.weakrefCallbackName)
        except AttributeError:
            pass


if __name__ == "__main__":
    import doctest
    doctest.testmod()
